#!/usr/bin/python3
"""
## Copyright University of Southampton IT Innovation Centre, 2018
##
## Copyright in this software belongs to University of Southampton
## IT Innovation Centre of Gamma House, Enterprise Road,
## Chilworth Science Park, Southampton, SO16 7NS, UK.
##
## This software may not be used, sold, licensed, transferred, copied
## or reproduced in whole or in part in any manner or form or in or
## on any media by any person other than in accordance with the terms
## of the Licence Agreement supplied with the software, or otherwise
## without the prior written consent of the copyright owners.
##
## This software is distributed WITHOUT ANY WARRANTY, without even the
## implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
## PURPOSE, except where stated in the Licence Agreement supplied with
## the software.
##
##      Created By :            Michael Boniface
##      Created Date :          15-04-2018
##      Updated By :            Nikolay Stanchev
##      Updated Date :          16-04-2018
##      Created for Project :   FLAME
"""


from influxdb import InfluxDBClient
import clmctest.monitoring.LineProtocolGenerator as lp
import urllib.parse
import time
import random


class Simulator(object):
    """
    Simulator used to generate E2E measurements.
    """

    DATABASE = 'CLMCMetrics'  # default database name
    DATABASE_URL = 'http://172.40.231.51:8086'  # default database url

    TICK = 1  # a simulation tick represents 1s
    SIMULATION_LENGTH = 120  # simulation time in seconds

    def __init__(self, database_url=DATABASE_URL, database=DATABASE):
        """
        Initialises the simulator by creating a db client object and resetting the database.

        :param database_url: db url
        :param database: db name
        """

        url_object = urllib.parse.urlparse(database_url)
        self.db_client = InfluxDBClient(host=url_object.hostname, port=url_object.port, database=database, timeout=10)

        self.db_url = database_url
        self.db_name = database

        self._reset_db()

    def _reset_db(self):
        """
        Reset the database using the already initialised db client object.
        """

        self.db_client.drop_database(self.db_name)
        self.db_client.create_database(self.db_name)

    def run(self):
        """
        Runs the simulation.
        """

        # all network delays start from 1ms, the dictionary stores the information to report
        paths = [
            {
                'target': 'SR2',
                'source': 'SR1',
                'path_id': 'SR1---SR2',
                'latency': 5,
                'bandwidth': 100*1024*1024
            },
            {
                'target': 'SR1',
                'source': 'SR2',
                'path_id': 'SR1---SR2',
                'latency': 5,
                'bandwidth': 100*1024*1024
            },
            {
                'target': 'SR3',
                'source': 'SR1',
                'path_id': 'SR1---SR3',
                'latency': 5,
                'bandwidth': 100*1024*1024
            },
            {
                'target': 'SR1',
                'source': 'SR3',
                'path_id': 'SR1---SR3',
                'latency': 5,
                'bandwidth': 100*1024*1024
            }
        ]

        service_function_instances = [
            {
                'endpoint': 'ms1.flame.org',
                'sf_instance': 'sr2.ms1.flame.org',  # TODO: what did we decide the sf_instance would look like?
                'sfr': 'SR2',
                'service_delay': 40,
                'cpus': 1
            },
            {
                'endpoint': 'ms1.flame.org',
                'sf_instance': 'sr3.ms1.flame.org',  # TODO: what did we decide the sf_instance would look like?
                'sfr': 'SR3',
                'service_delay': 10,
                'cpus': 4
            }
        ]

        av_request_size = 10 * 1024 * 1024  # average request size measured by service function / Bytes
        av_response_size = 1 * 1024  # average request size measured by service function / Bytes

        # current time in seconds (to test the aggregation we write influx data points related to future time), so we start from the current time
        start_time = int(time.time())

        sim_time = start_time

        sample_period_net = 1  # sample period for reporting network delays (measured in seconds)
        sample_period_media = 5  # sample period for reporting media service delays (measured in seconds)

        for i in range(0, self.SIMULATION_LENGTH):
            # report one of the network delays every sample_period_net seconds
            if i % sample_period_net == 0:
                path = random.choice(paths)
                self.db_client.write_points(
                    lp.generate_network_delay_report(path['path_id'], path['source'], path['target'], path['latency'], path['bandwidth'], sim_time))

                # increase/decrease the delay in every sample report (min delay is 1)
                path['latency'] = max(1, path['latency'] + random.randint(-3, 3))

            # report one of the service_function_instance response times every sample_period_media seconds
            if i % sample_period_media == 0:
                service = random.choice(service_function_instances)
                self.db_client.write_points(lp.generate_service_delay_report(
                    service['endpoint'], service['sf_instance'], service['sfr'], service['service_delay'], av_request_size, av_response_size, sim_time))

            # increase the time by one simulation tick
            sim_time += self.TICK

        end_time = sim_time
        print("Simulation finished. Start time: {0}, End time: {1}".format(start_time, end_time))


if __name__ == "__main__":
    Simulator().run()
